#ifdef HAVE_CONFIG_H
#include "config.h"
#endif 

#include <octave/oct.h>
#include <iostream>

#include "defun-dld.h"

#include "error.h"
#include "ov.h"
#include "variables.h"

#include "libdld_common.h"
#include "fifo_host.h"
#include "fileplace.h"
#include "process_info.h"

DEFUN_DLD (mpi_init, args, ,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} mpi_init ( @var{number} )\n\
\n\
Initialize the MPI execution environment\n\
\n\
@var{number}   - number of processes (scalar)\n\
@example\n\
@group\n\
\n\
For example,\n\
\n\
number_of_processes = 3;\n\
mpi_init( number_of_processes );\n\
...\n\
mpi_finalize();\n\
@end group\n\
@end example\n\
@end deftypefn")
{

  //Set directory-prefix where parallel-octave store files and named-pipe
  //We assume that this octave process is connected to rank0 mpi_process;
  Fileplace::set_prefix("0");

  if( args.length() != 1 )
    {
      print_usage("mpi_init");
      return octave_value();
    }

  if( poctave_common::poctave_is_active )
    return octave_value("octave is already invoked");

  pid_t pid;
  
  //Execute mpi process
  
  pid = fork();
  if(pid == -1)
    {
      perror("fork");
      exit(EXIT_FAILURE);
    }

  if( pid == 0 )
    {
      assert(setpgid(0,0)==0);
      char comm[1024];
      memset(comm,'\0',sizeof(comm));
      
      //LAM execute command
      
      int exec_size=args(0).int_value();
      char exec_size_c[512];
      memset(exec_size_c,'\0',sizeof(exec_size_c));
      sprintf(exec_size_c,"%d",exec_size);
      execlp("run_mpi.sh","run_mpi.sh","mpirun","-np",exec_size_c,NULL);
      
      perror("execvp");
      assert(1 == 0);
    }

  //Open IPC between this octave process and local MPI process

  char hostname[512];
  gethostname(hostname,sizeof(hostname));

  string fifopath1=Fileplace::getpath()+hostname+"magefifo";
  string fifopath2=Fileplace::getpath()+hostname+"hogefifo";

  poctave_common::mpiproc = new fifo_host(fifopath1,fifopath2,1);
  poctave_common::mpiproc->start();
  poctave_common::mpiproc->recv(&poctave_common::mpi_pid,
				    sizeof(poctave_common::mpi_pid));

  //Recv remote hostnum from local MPI process

  int remotehostnum;
  poctave_common::mpiproc->recv(&remotehostnum,sizeof(int));  

  for(int i = 0; i < remotehostnum; i++)
    {
      char recv_hostname[1024];
      int mpi_rank;
      int mpi_pid;
      poctave_common::mpiproc->recv(recv_hostname,sizeof(recv_hostname));
      poctave_common::mpiproc->recv(&mpi_rank,sizeof(int));
      poctave_common::mpiproc->recv(&mpi_pid,sizeof(int));
      
      node_process_info *p = new node_process_info;
      p->hostname = recv_hostname;
      p->mpi_rank = mpi_rank;
      p->mpi_pid = mpi_pid;
      process_info<node_process_info *>::instance()->push_back(p);
    }

  //Exec wrap

  for(process_info<node_process_info *>::iterator i=process_info<node_process_info *>::instance()->begin();
      i!=process_info<node_process_info *>::instance()->end();i++)
    {
      if( (*i)->mpi_rank != 0 )
	{
	  pid=fork();
	  if(pid==-1){
	    perror("fork");
	    exit(EXIT_FAILURE);
	  }
	  if(pid==0){
	    assert(freopen("/dev/null","r",stdin)!=NULL);
	    setsid();
	    char char_mpi_rank[512];
	    sprintf(char_mpi_rank,"%d",(*i)->mpi_rank);
	    char dir[512];
	    assert(getcwd(dir,sizeof(dir))!=NULL);
	    
	    execlp("run_wrap.sh",
		   "run_wrap.sh",
		   (*i)->hostname.c_str(), //remote hostname
		   "-s",
		   dir,
		   "-r",
		   char_mpi_rank,
		   "-h",
		   hostname,
		   NULL);
	    
	    perror("exec");
	  }
	}
    }
  
  // Accept connection from wrap process
  // -1 mean without this process
  poctave_common::host=new accept_host(23457,
				       process_info<node_process_info *>::instance()->size()-1);
  
  poctave_common::host->start();


  //Sync other node process start-up

  for(process_info<node_process_info *>::iterator i=process_info<node_process_info *>::instance()->begin();
      i!=process_info<node_process_info *>::instance()->end();i++)
    {
      if( (*i)->mpi_rank != 0 )
	{
	  //recieve remote node process information
	  string remote_hostname;
	  pid_t remote_pid;
	  poctave_common::host->scan_any_stream(remote_hostname,remote_pid,-1);
	  
	  pid_t octave_pid;
	  pid_t wrapper_pid;
	  int linked_mpi_rank;
	  
	  poctave_common::host->recvhost(remote_hostname,remote_pid,
					 &linked_mpi_rank,sizeof(int));
	  
	  poctave_common::host->recvhost(remote_hostname,remote_pid,
					 &octave_pid,sizeof(pid));

	  poctave_common::host->recvhost(remote_hostname,remote_pid,
					 &wrapper_pid,sizeof(pid));
	  
	  octave_stdout<<octave_pid<<","<<wrapper_pid<<","<<
	    linked_mpi_rank<<","<<(*i)->hostname.c_str()<<std::endl;
	  
	  //check connectting octave process is same as linked mpi-process
	  //this should be removed.
	  process_info<node_process_info *>::iterator j;
	  bool not_found=true;
	  for(j=process_info<node_process_info *>::instance()->begin();
	      j!=process_info<node_process_info *>::instance()->end();j++){
	    if((*j)->mpi_rank==linked_mpi_rank){
	      not_found=false;
	      (*j)->octave_pid=octave_pid;
	      (*j)->wrapper_local_to_remote_pid=wrapper_pid;
	    }
	  }
	  if(not_found)
	    assert(false);

	}
    }
  
  poctave_common::poctave_is_active=true;

  return octave_value("poctave invoked");

};
