#include <octave/oct.h>
#include <octave/byte-swap.h>

#include "libdld_common.h"
#include "fifo_host.h"

DEFUN_DLD(mpi_recv, args, ,
	  "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {[@var{msg}]=} mpi_recv ( @var{size}, @var{src}, @var{tag}, @var{comm} )\n\
Basic receive.\n\
\n\
INPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{size}\n   size of octave_value_list to receive (scalar)\n\
@item @var{src}\n    rank of source (scalar)\n\
@item @var{tag}\n    message tag (scalar)\n\
@item @var{comm}\n   communicator (scalar)\n\
@end table\n\
\n\
OUTPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{msg}\n    received message (octave_value)\n\
@end table\n\
\n\
@example\n\
@group\n\
\n\
For example,\n\
\n\
tag = 0; dest = 0; src = 0;\n\
mpi_send ( \"hello world\", sizeof(\"hello world\"), dest, tag, MPI_COMM_WORLD );\n\
mpi_recv ( sizeof(\"hello world\"), src, tag, MPI_COMM_WORLD )\n\
\n\
@end group\n\
@end example\n\
@end deftypefn")
{
  if( args.length() != 4 )
    {
      print_usage("mpi_recv");
      return octave_value();
    }

  if( ! poctave_common::poctave_is_active )
    return octave_value("mpi_init must be called beforehand");
  
  int size;
  int src;
  int tag;
  int key; //COMMUNICATOR

  if( args(0).is_real_scalar() &&
      args(1).is_real_scalar() &&
      args(2).is_real_scalar() &&
      args(3).is_real_scalar())
    {
      size=args(0).int_value();
      src =args(1).int_value();
      tag =args(2).int_value();
      key =args(3).int_value();
    }
  else
    {
      error("mpi_recv: bad reciever");
      return octave_value();
    }

  fifo_host *mpiproc=(fifo_host *)poctave_common::mpiproc;
  
  try
    {
      mpiproc->send("mpi_recv",sizeof("mpi_recv"));
      
      mpiproc->send(&size,sizeof(int));
      mpiproc->send(&src,sizeof(int));
      mpiproc->send(&tag,sizeof(int));
      mpiproc->send(&key,sizeof(int));
      
      int buffer_length;
      mpiproc->recv(&buffer_length,sizeof(int));
      char buffer_data[buffer_length];
      mpiproc->recv(buffer_data,buffer_length);
      
      // Pull the entry _start and _length data off the end of the buffer
      std::basic_stringstream<char> recv_buffer;
      std::basic_streambuf<char> *rdbuf;
      
      rdbuf = recv_buffer.rdbuf();
      rdbuf->pubsetbuf(buffer_data, buffer_length);
      
      // Read the float format
      oct_mach_info::float_format flt_fmt;
      
      recv_buffer.read(X_CAST(char *,&flt_fmt),
		       sizeof(oct_mach_info::float_format));
      
      // Convert binary buffer_data back to an octave value
      octave_value_list retr_list;
      bool global;
      octave_value recv_value;
      std::string help;
      std::string var_name;
      
      while(!recv_buffer.eof())
	{
	  var_name = mpi_read_binary_data (recv_buffer, false, 
					   flt_fmt,
					   "", global, recv_value, help);
	  if(var_name.empty())
	    break;
	  else
	    retr_list.append(recv_value);
	  
	}

    return retr_list;
    }
  catch(std::string e)
    {
      kill(poctave_common::mpi_pid,SIGURG);
      perror(e.c_str());
      return octave_value(e);
    }

}
