#include <octave/oct.h>
#include <octave/byte-swap.h>

#include "libdld_common.h"
#include "fifo_host.h"

DEFUN_DLD(mpi_send, args,,
	  "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {[@var{size}]=} mpi_send ( @var{msg}, @var{size}, @var{dest}, @var{tag}, @var{comm} )\n\
Performs a basic send\n\
\n\
INPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{msg}\n    ocatve_value to send (octave_value)\n\
@item @var{size}\n   size of octave_value to send (scalar)\n\
@item @var{dest}\n   rank of destination (scalar)\n\
@item @var{tag}\n    message tag (scalar)\n\
@item @var{comm}\n   communicator (scalar)\n\
@end table\n\
\n\
OUTPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{size}\n   the number of characters sent (scalar)\n\
@end table\n\
\n\
@example\n\
@group\n\
\n\
For example,\n\
\n\
tag = 0; dest = 0; src = 0;\n\
mpi_send ( \"hello world\", sizeof(\"hello world\"), dest, tag, MPI_COMM_WORLD );\n\
mpi_recv ( sizeof(\"hello world\"), src, tag, MPI_COMM_WORLD )\n\
\n\
@end group\n\
@end example\n\
@end deftypefn")
{
  if ( args.length() < 5 )
    {
      print_usage("mpi_send");
      return octave_value();
    }
  
  if ( ! poctave_common::poctave_is_active )
      return octave_value("mpi_init must be called beforehand");

  int size;
  int dest;
  int tag;
  int key; //COMMUNICATOR;
  
  // Parse args
  if ( args(args.length()-1).is_real_scalar() &&
       args(args.length()-2).is_real_scalar() &&
       args(args.length()-3).is_real_scalar() &&
       args(args.length()-4).is_real_scalar() )
    {
      size=args(args.length()-4).int_value();
      dest=args(args.length()-3).int_value();
      tag =args(args.length()-2).int_value();
      key =args(args.length()-1).int_value();
    }
  else
    {
      error("mpi_send: bad operant");
      return octave_value();
    }

  std::basic_ostringstream<char> send_buffer;

  // Save the byte format to the stream
  oct_mach_info::float_format flt_fmt = 
    oct_mach_info::native_float_format();

  send_buffer.write(X_CAST(char *,&flt_fmt),
		    sizeof(oct_mach_info::float_format));
		    
  // Serialize all each entry in binary form
  for (int i = 0; i< args.length()-4 ; i++)
    {
      // Make sure the value exists
      if (args(i).is_defined())
	{
	  // Convert octave_value to binary
	  mpi_save_binary_data(send_buffer, 
			       args(i), "V", "", 
			       false, false);
	}
      else
	{
	  error("mpi_send: Doesn't handle that form of data yet");
	  
	  return octave_value();
	}
    }
  
  send_buffer.eof();
  
  std::basic_string<char> send_data = send_buffer.str();
  int buffer_length = send_data.length();
  const char *buffer_data = send_data.data();
  
  int retval=buffer_length;

  try
    {
      fifo_host *mpiproc=(fifo_host *)poctave_common::mpiproc;
      
      mpiproc->send("mpi_send",sizeof("mpi_send"));

      mpiproc->send(&size,sizeof(int));

      mpiproc->send(&buffer_length,sizeof(int));
      mpiproc->send(buffer_data,buffer_length);
      
      mpiproc->send(&dest,sizeof(int));
      mpiproc->send(&tag,sizeof(int));
      mpiproc->send(&key,sizeof(int));

      mpiproc->recv(&retval,sizeof(int));

    }
  catch(std::string e)
    {
      kill(poctave_common::mpi_pid,SIGURG);
      return octave_value("cancel");
    }
  
  return octave_value(buffer_length);
}


