#include <octave/oct.h>
#include <defun-dld.h>
#include <error.h>
#include <ov.h>

#include "libdld_common.h"
#include "fifo_host.h"

DEFUN_DLD (mpi_wait, args, ,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} { [ @var{msg} @var{stat} ] =} mpi_wait ( @var{req} )\n\
Waits for an MPI send or receive to complete\n\
\n\
INPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{req}\n communication request (scalar)\n\
@end table\n\
\n\
OUTPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{msg}\n received message is set if @var{stat} is true. If @var{stat} is faise, error message is returned. (octave_value)\n\
@item @var{stat}\n If an error occures, false is returned. otherwise true is returned (scalar) \n\
@end table\n\
\n\
@example\n\
@group\n\
\n\
For example,\n\
\n\
tag = 0; dest = 0; src = 0;\n\
req = mpi_irecv ( sizeof(\"hello world\"), src, tag, MPI_COMM_WORLD )\n\
if ( mpi_test( req ) )\n\
\t [msg stat] = mpi_wait( req )\n\
\n\
@end group\n\
@end example\n\
@end deftypefn")
{
  octave_value_list retval;

  if( args.length() != 1 )
    {
      print_usage("mpi_wait");
      retval(0) = std::string("invalid argument");
      retval(1) = octave_value(false);
      return retval;
    }

  if( ! poctave_common::poctave_is_active )
    {
      retval(0) = std::string("mpi_init must be called beforehand.");
      retval(1) = octave_value(false);
      return retval;
    }
  
  int key; //Request

  if( args(0).is_real_scalar() ){
    key = args(0).int_value();
  }
  else
    {
      error("request must be scalar");
      retval(0) = std::string("request must be scalar");
      retval(1) = false;
      return retval;
    }
  
  try
    {
      fifo_host *mpiproc = (fifo_host *)poctave_common::mpiproc;
      
      mpiproc->send("mpi_wait",sizeof("mpi_wait"));
      
      mpiproc->send(&key,sizeof(int));
      
      int mpi_retval;
      mpiproc->recv(&mpi_retval,sizeof(int));
      if(mpi_retval == -1)
	{
	  retval(0)=std::string("error in mpi-module");
	  retval(1)=false;
	  return retval;
	}
      
      int buffer_length;
      mpiproc->recv(&buffer_length,sizeof(int));
      char buffer_data[buffer_length];
      mpiproc->recv(buffer_data,buffer_length);
      
      // Pull the entry _start and _length data off the end of the buffer
      std::basic_stringstream<char> recv_buffer;
      std::basic_streambuf<char> *rdbuf;
      
      rdbuf = recv_buffer.rdbuf();
      rdbuf->pubsetbuf(buffer_data, buffer_length);
      
      // Read the float format
      oct_mach_info::float_format flt_fmt;
      
      recv_buffer.read(X_CAST(char *,&flt_fmt),
                       sizeof(oct_mach_info::float_format));
      
      // Convert binary buffer_data back to an octave value
      bool global;
      octave_value recv_value;
      std::string help;
      std::string var_name;

      while(!recv_buffer.eof())
        {
          var_name = mpi_read_binary_data (recv_buffer, false,
                                           flt_fmt,
                                           "", global, recv_value, help);
          if(var_name.empty())
            break;
          else
            retval.append(recv_value);
	  
        }

      retval.append(octave_value(true));
      return retval;
    }
  catch(std::string e)
    {
      kill(poctave_common::mpi_pid,SIGURG);
      retval(0) = std::string("operation canceled");
      retval(1) = false;
      return retval;
    }
  
}




