#include <octave/oct.h>
#include <defun-dld.h>
#include <error.h>
#include <ov.h>

#include "libdld_common.h"
#include "fifo_host.h"

DEFUN_DLD (mpi_waitall, args, nargout,
           "-*- texinfo -*-\n\
@deftypefn {Loadable Function} { [ @var{stat} @var{msgs}... ] =} mpi_testall ( @var{reqs} )\n\
Waits for all given communications to complete\n\
\n\
INPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{reqs}\n communication requests (matrix)\n\
@end table\n\
\n\
OUTPUT PARAMETERS\n\
\n\
@table @code\n\
@item @var{stat}\n true only if all requests have completed. If operation is not completed, or an error occures, false is returned. (bool)\n\
@item @var{msgs}\n received messages are set if @var{stat} is true. If @var{stat} is faise, error message is set to first msgs. And other msgs are set 0 (octave_value)\n\
@end table\n\
\n\
@example\n\
@group\n\
\n\
For example,\n\
\n\
tag1 = 0; tag2 = 1; dest = 0; src = 0;\n\
msg1 = \"hello world with tag1\";\n\
msg2 = \"hello world with tag2\";\n\
req1 = mpi_irecv ( sizeof(msg1), src, tag1, MPI_COMM_WORLD )\n\
req2 = mpi_irecv ( sizeof(msg2), src, tag2, MPI_COMM_WORLD )\n\
mpi_send ( msg1,  sizeof(msg1), dest, tag1, MPI_COMM_WORLD )\n\
mpi_send ( msg2,  sizeof(msg2), dest, tag2, MPI_COMM_WORLD )\n\
[ stat rev_msg1 rev_msg2 ] = mpi_waitall( [ req1 req2 ] );\n\
if ( stat )\n\
  printf(\"%s\\n%s\\n\",rev_msg1,rev_msg2);\n\
else\n\
  error(rev_msg1);\n\
endif\n\
\n\
@end group\n\
@end example\n\
@end deftypefn")
{
  octave_value_list retval;

  if( args.length() != 1 )
    {
      print_usage("mpi_waitall");
      return retval;
    }
  
  if( ! poctave_common::poctave_is_active )
    {
      error("mpi_init must be called beforehand.");
      return retval;
    }

  Matrix key; //requests

  if ( args(0).is_scalar_type() || args(0).is_real_matrix() )
    {
      key = args(0).matrix_value(); 
      if( key.rows() != 1 )
	{
	  error ("rows of reqs must be 1 ");
	  return retval;	
	}
    }
  else
    {
      error ("reqs must be scalar or matrix");
      return retval;
    }
    
  fifo_host *mpiproc=(fifo_host *)poctave_common::mpiproc;

  mpiproc->send("mpi_waitall",sizeof("mpi_waitall"));

  int col=key.columns();
  mpiproc->send(&col,sizeof(int));

  for(int i = 0; i < col; i++)
    {
      int tmp=int(key(0,i));
      mpiproc->send(&tmp,sizeof(int));
    }
  
  int mpi_retval;

  mpiproc->recv(&mpi_retval,sizeof(int));

  retval.append(octave_value(mpi_retval));

  if( mpi_retval > 0 )
    {
      for(int i = 0 ; i < col ;i++ )
	{
	  int buffer_length;
	  mpiproc->recv(&buffer_length,sizeof(int));
	  char buffer_data[buffer_length];
	  mpiproc->recv(buffer_data,buffer_length);
	  
	  // Pull the entry _start and _length data off the end of the buffer
	  std::basic_stringstream<char> recv_buffer;
	  std::basic_streambuf<char> *rdbuf;
	  
	  rdbuf = recv_buffer.rdbuf();
	  rdbuf->pubsetbuf(buffer_data, buffer_length);
	  
	  // Read the float format
	  oct_mach_info::float_format flt_fmt;
	  
	  recv_buffer.read(X_CAST(char *,&flt_fmt),
			   sizeof(oct_mach_info::float_format));
	  
	  // Convert binary buffer_data back to an octave value
	  bool global;
	  octave_value recv_value;
	  std::string help;
	  std::string var_name;
	  
	  while( ! recv_buffer.eof() )
	    {
	      var_name = mpi_read_binary_data (recv_buffer, false,
					       flt_fmt,
					       "", global, recv_value, help);
	      if(var_name.empty())
		break;
	      else
		retval.append(recv_value);	  
	    }
	}
    }
  else
    {
      for( int i = 1; i < nargout; i++ )
	retval.append(octave_value(0));
    }

  return retval;

}


